<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <link rel="icon" type="image/x-icon" href="/cf931f8d6f928f83280b.ico">
    <link rel="stylesheet" href="/styles.css">
    <title><%= htmlWebpackPlugin.options.title %></title>
</head>
<body>
    <section class="section">
        <div class="container is-flex is-justify-content-center is-align-items-center">
            <article class="message is-danger">
                <div class="message-header">
                  <p>404 Not Found</p>
                </div>
                <div class="message-body">
                  Sorry, we couldn't find what you were looking for. Click <a href="/"><strong>here</strong></a> to return to the home page.
                </div>
              </article>
        </div>
    </section>
</body>
</html>